"""Runtime model representing the plugged state of the device."""

from .uphy import Signal, Parameter
from typing import Optional, List
from pydantic import BaseModel


class Signal(Signal):
    """Signal datatype."""

    description: Optional[str]
    id: Optional[str]


class Parameter(Parameter):
    """Parameter datatype."""

    description: Optional[str]
    id: Optional[str]


class Slot(BaseModel):
    """Slot datatype."""

    name: str
    inputs: List[Signal]
    outputs: List[Signal]
    parameters: List[Parameter]


class Device(BaseModel):
    """Device datatype."""

    name: str
    slots: List[Slot]


class Root(BaseModel):
    """Root datatype."""

    device: Device



from enum import Enum
from typing import Optional, List
from pydantic import BaseModel, Field, UUID4, PrivateAttr


class MBox(BaseModel):
    offset: int
    size: int


class Fmmu(Enum):
    Unused = 0
    Outputs = 1
    Inputs = 2
    MBoxState = 3


class SMType(Enum):
    MBoxOut = 1
    MBoxIn = 2
    Outputs = 3
    Inputs = 4


class SM(BaseModel):
    address: int
    length: int
    control: int
    status: int
    enable: int
    type: SMType


class CoE(Enum):
    Sdo = 1 << 0
    SdoInfo = 1 << 1
    PdoAssign = 1 << 2
    PdoConfiguration = 1 << 3
    PdoUpload = 1 << 4
    CompleteAccess = 1 << 5

class FoE(Enum):
    Enable = 1 << 0

class MBoxProtocol(Enum):
    EoE = 1 << 1
    CoE = 1 << 2
    FoE = 1 << 3
    SoE = 1 << 4
    VoE = 1 << 5


class CiAObject(BaseModel):
    """EtherCAT process data"""
    index: str = Field(title="Index")
    subindex: str = Field(title="Subindex")
    signal: UUID4 = Field(title="Signal")


class CiAPDO(BaseModel):
    """An EtherCAT Rx or Tx PDO. An Rx PDO is an output signal, a Tx PDO is an input signal."""
    name: str = Field(title="PDO name")
    index: str = Field(title="Index")
    entries: List[CiAObject] = Field(title="PDO entries", default=[])


class EtherCATModule(BaseModel):
    """EtherCAT specific module configuration. Maps signals to Rx or Tx PDO:s"""
    rxpdo: List[CiAPDO] = Field(title="Rx PDO", default=[])
    txpdo: List[CiAPDO] = Field(title="Tx PDO", default=[])
    objects: List[CiAObject] = Field(title="SDO", default=[])
    profile: str = Field(title="EtherCAT profile", default="5001")

    def generate_mapping(self, module, module_ix):
        if len(module.outputs) > 0 and len(self.rxpdo) == 0:
            pdo = CiAPDO(
                name = "Outputs",
                index = "#x1600",
            )
            for ix, signal in enumerate(module.outputs):
                entry = CiAObject(
                    index = f"#x{0x6000 + ix:x}",
                    subindex = 0,
                    signal = signal.id
                )
                pdo.entries.append(entry)
            self.rxpdo.append(pdo)

        if len(module.inputs) > 0 and len(self.txpdo) == 0:
            pdo = CiAPDO(
                name = "Inputs",
                index = "#x1A00",
            )
            for ix, signal in enumerate(module.inputs):
                entry = CiAObject(
                    index = f"#x{0x7000 + ix:x}",
                    subindex = 0,
                    signal = signal.id
                )
                pdo.entries.append(entry)
            self.txpdo.append(pdo)

        if len(module.parameters) > 0 and len(self.objects) == 0:
            for ix, parameter in enumerate(module.parameters):
                entry = CiAObject(
                    index = f"#x{0x8000 + ix:x}",
                    subindex = 0,
                    signal = parameter.id
                )
                self.objects.append(entry)

class EtherCATDevice(BaseModel):
    """EtherCAT specific device configuration"""
    product_code: str = Field(title="Product code", default="0")
    revision: str = Field(title="Revision", default="0")
    profile: str = Field(title="EtherCAT profile", default="5001")


class EtherCATRoot(BaseModel):
    """EtherCAT config"""
    vendor_id: str = Field(title="Vendor ID")
    group: str = Field(title="Group", default="U-Phy")

    _pdi_control: int = PrivateAttr(default=0x0e82)  # SPI EtherCAT Direct Mapped Mode
    _pdi_configuration: int = PrivateAttr(default=0x0681)
    _asic_config: int = PrivateAttr(default=0x4000)  # ERRLED enable
    _coe: int = PrivateAttr(default=(CoE.Sdo.value | CoE.SdoInfo.value))
    _foe: int = PrivateAttr(default=FoE.Enable.value)
    _mbox_control: int = PrivateAttr(default=(
        MBoxProtocol.CoE.value |
        MBoxProtocol.FoE.value
    ))
    _eeprom_size: int = PrivateAttr(512)  # bytes
    _boot_recv_mbox: MBox = PrivateAttr(default=MBox(
        offset=0x1000,
        size=512
    ))
    _boot_send_mbox: MBox = PrivateAttr(default=MBox(
        offset=0x1200,
        size=512
    ))
    _standard_recv_mbox: MBox = PrivateAttr(default=MBox(
        offset=0x1000,
        size=512
    ))
    _standard_send_mbox: MBox = PrivateAttr(default=MBox(
        offset=0x1200,
        size=512
    ))
    _fmmu: List[Fmmu] = PrivateAttr(default=[
        Fmmu.Outputs,
        Fmmu.Inputs,
        Fmmu.MBoxState,
    ])
    _sm: List[SM] = PrivateAttr(default=[
        SM(
            address=0x1000,
            length=512,
            control=0x26,
            status=0,
            enable=1,
            type=SMType.MBoxOut,
        ),
        SM(
            address=0x1200,
            length=512,
            control=0x22,
            status=0,
            enable=1,
            type=SMType.MBoxIn,
        ),
        SM(
            address=0x1400,
            length=0,
            control=0x64,
            status=0,
            enable=1,
            type=SMType.Outputs,
        ),
        SM(
            address=0x1A00,
            length=0,
            control=0x20,
            status=0,
            enable=1,
            type=SMType.Inputs,
        ),
    ])



import logging
import re
import json
import importlib.resources
from .. import templates
from jinja2 import Environment, PackageLoader, select_autoescape
from ..model.uphy import *

_logger = logging.getLogger(__name__)


def c_type(value):
    map = {
        DataType.UINT32: 'uint32_t',
        DataType.UINT16: 'uint16_t',
        DataType.UINT8: 'uint8_t',
        DataType.INT32: 'int32_t',
        DataType.INT16: 'int16_t',
        DataType.INT8: 'int8_t',
        DataType.REAL32: 'float',
    }
    return map[value.datatype]

class CNameContext():
    """
    Defines scopes for identifiers:
    - Slots
        - Signals/Parameters (per slot)
    - Modules
    - Alarms
    """
    class Scope(Enum):
        NONE = 0,
        SLOT = 1,
        SIGNAL = 2,
        MODULE = 3,
        ALARM = 4,
        TXPDO = 5,
        RXPDO = 6

    def __init__(self, slot=""):
        self.slot = slot
        self.scope = self.Scope.NONE
        self.ctx = {
            self.Scope.SLOT: {},    # (name, index)
            self.Scope.SIGNAL: {},  # {slot_index: (name, index)}
            self.Scope.MODULE: {},  # (name, index)
            self.Scope.ALARM: {},   # (name, index)
            self.Scope.TXPDO: {},   # (name, index)
            self.Scope.RXPDO: {},   # (name, index)
        }

    def _get_ctx(self):
        assert self.ctx != self.Scope.NONE

        if self.Scope == self.Scope.SIGNAL:
            assert self.slot != ""
            if self.slot not in self.ctx[self.scope]:
                self.ctx[self.scope][self.slot] = {}

            ctx = self.ctx[self.scope][self.slot]
        else:
            ctx = self.ctx[self.scope]

        return ctx

    def name_available(self, name, index):
        ctx = self._get_ctx()

        for other_index, other_name in ctx.items():
            if name == other_name and index != other_index:
                _logger.debug(f"[{self.scope}] name collision: {name}, '{index}' '{other_index}'")
                return False

        return True

    def set_name(self, name, index):
        ctx = self._get_ctx()
        ctx[index] = name

    def get_name(self, index):
        ctx = self._get_ctx()
        return ctx.get(index, None)

    def index_exists(self, index):
        ctx = self._get_ctx()
        return index in ctx

    def new_unique_name(self, name, index):
        if self.index_exists(index):
            name = self.get_name(index)
        else:
            while not self.name_available(name, index):
                name += "_" + str(index)

            # Store name for future calls
            self.set_name(name, index)

            _logger.debug(f"c_name: {name} allocated to index '{index}'")

        return name

def c_name(value, index, ctx, scope):
    """
    Filter for C names. An index should be included to avoid duplicate
    names and ensure consistency across repeated calls. For example:
        {{slot | c_name(loop.index0, ctx, ctx.Scope.SLOT)}}
        {{signal | c_name("in_" ~ loop.index0, ctx, ctx.Scope.SIGNAL)}}

    Valid scopes are defined by the enum ctx.Scope (except NONE):
    - ctx.Scope.SLOT
        - ctx.Scope.SIGNAL (NOTE: requires previous call with scope SLOT)
    - ctx.Scope.MODULE
    - ctx.Scope.ALARM
    - ctx.Scope.TXPDO
    - ctx.Scope.RXPDO

    Note that the SIGNAL scope is shared by inputs, outputs and parameters.
    """
    name = value.name
    if re.search("^[0-9].*", name):           # Escape leading numeral
        name = f"_{name}"

    name = re.sub("[^0-9A-Za-z_]", "_", name) # Replace special chars
    name = re.sub("_+", "_", name)            # Contract repeating underscore
    if name != "_":
        name = re.sub("_$", "", name)         # Strip trailing underscore

    # Set scope
    ctx.scope = scope
    if scope == ctx.Scope.SLOT:
        ctx.slot = index                      # Define signal scope

    name = ctx.new_unique_name(name, index)   # Ensure name is unique

    return name


def c_name_upper(value, index, ctx, scope):
    """Filter for uppercase C names. For example defines."""
    return c_name(value, index, ctx, scope).upper()


def c_hex(value):
    return value.replace('#', '0')

def c_array(signal):
    if not signal.is_array:
        return ""
    else:
        return f"[{signal.array_length}]"

def c_flags(signal):
    if not signal.is_array:
        return "0"
    else:
        return "UP_SIG_FLAG_IS_ARRAY"

def c_bool(value):
    if value:
        return "true"
    else:
        return "false"

# str.__repr__ never escapes double quotes:
# - print(("\"").__repr__())        => '"'
# - print(("\"" + '\'').__repr__()) => '\'"'
# Backslashes also need to appear escaped when output.
def c_string(s):
    if s:
        s = s.replace("\\", "\\\\")
        s = s.replace('"', r'\"')
    return s

# str.__repr__ only escapes single quotes if there is also a
# double quote in the string:
# - print(('\'').__repr__())        => "'"
# - print(("\"" + '\'').__repr__()) => '\'"'
# For C chars an extra backslash therefore needs to be added.
def c_char(s):
    if s:
        s = s.replace("'", r"\'")
    return s

def enip_major_rev(rev_str):
    return rev_str.split('.')[0]

def enip_minor_rev(rev_str):
    return rev_str.split('.')[1]

class CCodeGenerator:
    def __init__(self, model):
        self.model = model
        self.device = None
        self.name_ctx = CNameContext()

        self.env = Environment(
            loader=PackageLoader("upgen"),
            autoescape=select_autoescape(),
            trim_blocks=True,
            lstrip_blocks=True
        )
        self.env.filters['c_type'] = c_type
        self.env.filters['c_name'] = c_name
        self.env.filters['c_hex'] = c_hex
        self.env.filters['c_name_upper'] = c_name_upper
        self.env.filters['c_array'] = c_array
        self.env.filters['c_flags'] = c_flags
        self.env.filters['c_bool'] = c_bool
        self.env.filters['c_string'] = c_string
        self.env.filters['c_char'] = c_char
        self.env.filters['enip_major_rev'] = enip_major_rev
        self.env.filters['enip_minor_rev'] = enip_minor_rev
        self.env.globals['ctx'] = self.name_ctx

    def select_device(self, device_name):
        if device_name == None:
            self.device = self.model.devices[0]
        else:
            self.device = self.model.get_device(device_name)

        if self.device == None:
            raise Exception(f'{device_name} not found in model')

    def export_file(self):
        _logger.info('Generate model.h')
        header = self.generate_header_file()
        _logger.info('Generate model.c')
        source = self.generate_source_file()

        with open('model.h', 'w') as file:
            file.write(header)

        with open('model.c', 'w') as file:
            file.write(source)

    def generate_header_file(self):
        template = self.env.get_template("model_template.h")
        code = template.render(model=self.model, device=self.device)
        _logger.debug(code)
        return code

    def generate_source_file(self):
        template = self.env.get_template("model_template.c")
        code = template.render(model=self.model, device=self.device)
        _logger.debug(code)
        return code



import logging
from pathlib import Path
import os
import re

import importlib.resources
from .. import templates
from jinja2 import Environment, PackageLoader, select_autoescape
from ..model.uphy import *
from datetime import date, datetime

_logger = logging.getLogger(__name__)

eip_type = {
    'UINT32': '0xC8',
    'UINT16': '0xC7',
    'UINT8': '0xC6',
    'INT32': '0xC4',
    'INT16': '0xC3',
    'INT8': '0xC2',
    'REAL32':'0xCA',
}

eip_bytelen = {
    'UINT32': 4,
    'UINT16': 2,
    'UINT8': 1,
    'INT32': 4,
    'INT16': 2,
    'INT8': 1,
    'REAL32': 4,
}

class EipAssemblyElement():
    def __init__(self, eds_ref, size_in_bytes : int):
        self.eds_ref = eds_ref
        self.size_in_bytes = size_in_bytes
        self.size_in_bits = size_in_bytes*8

class EipParam():
    id: int = 1
    objects = []

    def get_id():
        id = EipParam.id
        EipParam.id = EipParam.id + 1
        return id

    def reset():
        '''
        Reset class variables. Needed for automatic testing.
        '''
        EipParam.id = 1
        EipParam.objects = []

    def create_params(name = "None", signal = None,
                 data_type = None, value_min = "",
                 value_max = "", value_default = "0", description=""):
        '''
        Prefered method to create parameters.
        To handle array signals a list of parameters is returned.
        An array signal with length n generates n parameters. For scalar signals the length of the list will be 1.
        '''
        params = []

        if signal != None:
            if type(signal) == Signal:
                if signal.is_array:
                    for i in range(0, signal.array_length):
                        array_name = f'{name}[{i}]'
                        p = EipParam(array_name, signal.datatype, help = signal.description, model_ref=signal.id)
                        params.append(p)
                else:
                    p = EipParam(name, signal.datatype, help = signal.description, model_ref=signal.id)
                    params.append(p)
            elif type(signal) == Parameter:
                p = EipParam(name, signal.datatype, help = signal.description, model_ref=signal.id)
                params.append(p)
        else:
            p = EipParam(name, data_type, value_min, value_max, value_default, description)
            params.append(p)
        return params

    def __init__(self, name,
                 data_type = None, value_min = "",
                 value_max = "", value_default = "0", help="", model_ref=None):

        self.id: int = EipParam.get_id()
        self.eds_ref = f'Param{self.id}'
        self.name = name
        self.help = help
        self.model_ref = model_ref
        self.data_type = data_type
        self.value_min = value_min
        self.value_max = value_max
        self.value_default = value_default
        self.size = 0
        self.eds_data_type = eip_type[self.data_type]
        if self.data_type != None:
            self.size = eip_bytelen[self.data_type]
        self.descriptor = 0
        EipParam.objects.append(self)
        _logger.info(f'Create {self.eds_ref:12} {self.name:64} {self.data_type} {self.value_default}')

    def get_size(self):
        return self.size

    def to_eds_assembly_element(self):
        return EipAssemblyElement(self.eds_ref, self.get_size())

class EipAssembly():
    id: int = 100
    objects = []

    def get_id():
        id = EipAssembly.id
        EipAssembly.id = EipAssembly.id + 1
        return id

    def reset():
        '''
        Reset class variables. Needed for automatic testing.
        '''
        EipAssembly.id = 100
        EipAssembly.objects = []

    def remove_unused_assemblies():
        '''
        Remove assembly objects with size == 0 from the Assembly objects list
        This should only apply for input, output or config assemblies in the case
        of a device does not have inputs, outputs or configuration parameters
        Note that assemblies are only removes from the list, which is used to
        generate the AssemN objects in [Assembly] section.
        The assembly ids are still kept to create unique connection paths and
        to avoid special cases in the corresponding enip adapter code.
        '''
        EipAssembly.objects = [o for o in EipAssembly.objects if o.get_size() > 0]


    def __init__(self, name:str, signal_list: BaseSignal = None, param_prefix: str = None):
        '''
        Create assembly object. If a list of signal is passed as argument, then parameters
        are created for all signals in the list and the parameters are appended to the assembly.
        The param_prefix is used to create unique parameter names.
        '''
        self.id: int = EipAssembly.get_id()
        self.eds_ref = f'Assem{self.id}'
        EipAssembly.objects.append(self)
        self.name = name
        self.subelements = []

        _logger.info(f'Create {self.eds_ref:12} {self.name} (param prefix {param_prefix})')

        if signal_list != None:
            for signal in signal_list:
                if (param_prefix != None):
                    param_name = f'{param_prefix}.{signal.name}'
                    params = EipParam.create_params(name = param_name, signal=signal)
                else:
                    params = EipParam.create_params(signal=signal)

                for p in params:
                    self.append_element(p.to_eds_assembly_element())

    def get_size(self):
        size = 0
        for e in self.subelements:
            size = size + e.size_in_bytes
        return size

    def append_element(self, element: EipAssemblyElement):
        self.subelements.append(element)

    def to_eds_assembly_element(self):
        return EipAssemblyElement(self.eds_ref, self.get_size())

    def id_hex_str(self):
        return f'{self.id:0>2X}'


def assembly_elements_to_eds(elements):
    txt = '                ,,\n'
    for element in elements:
        txt = f"{txt}                {element.size_in_bits},{element.eds_ref},\n"

    txt = f"{txt[:-2]};" # replace final ',\n' with ';'
    return txt


# Use timestamp if available. If not, use current date
def get_date(date):
    if bool(date):
        return date
    else:
        return datetime.now().strftime("%m-%d-%Y")

# Use timestamp if available. If not, use current time
def get_time(time):
    if bool(time):
        return time
    else:
        return datetime.now().strftime("%H:%M:%S")

def enip_major_rev(rev_str):
    return rev_str.split('.')[0]

def enip_minor_rev(rev_str):
    return rev_str.split('.')[1]

class EthernetipGenerator:
    def __init__(self, model):
        self.model = model
        self.eds_file_name = f"{model.name}.eds"
        self.device = None
        self.connections = []               # Connections
        self.module_to_assembly_map = []
        self.rpi_param = None

        self.env = Environment(
            loader=PackageLoader("upgen"),
            autoescape=select_autoescape(),
            trim_blocks=True,
            lstrip_blocks=True
        )

        self.env.filters['assembly_elements_to_eds'] = assembly_elements_to_eds
        self.env.filters['date'] = get_date
        self.env.filters['time'] = get_time
        self.env.filters['enip_major_rev'] = enip_major_rev
        self.env.filters['enip_minor_rev'] = enip_minor_rev

    def select_device(self, device_name):
        if device_name == None:
            self.device = self.model.devices[0]
        else:
            self.device = self.model.get_device(device_name)
        if self.device == None:
            raise Exception(f'{device_name} not found in model')

        self.eds_file_name = f"{self.device.name}.eds"

    def export_file(self):
        if Adapter.ETHERNETIP not in self.model.adapters:
            _logger.debug("EtherNet/IP is not enabled")
            return

        if not self.model.ethernetip:
            _logger.warn('Model does not include a valid EtherNet/IP configuration')
            return

        _logger.info('Generate eds file')
        eds = self.generate_eds()

        _logger.info(f"Save EDS file: {str(Path(os.getcwd()) / self.eds_file_name)}")
        with open(self.eds_file_name, 'w') as file:
            file.write(eds)

    def create_connection_dict(self, eds_ref, name, trigger_and_transport, connection_parameters, help, rpi_param, input_assembly, output_assembly, config_assembly,path):
        connection = {
            'eds_ref' : eds_ref,
            'name' : name,
            'trigger_and_transport' : trigger_and_transport,
            'connection_parameters' : connection_parameters,
            'help' : help,
            'input_rpi_param' : rpi_param.eds_ref,
            'input_size' : input_assembly.get_size(),
            'input_format' : input_assembly.eds_ref,
            'output_rpi_param' : rpi_param.eds_ref,
            'output_size' : output_assembly.get_size(),
            'output_format' : output_assembly.eds_ref,
            'config_size' : config_assembly.get_size(),
            'config_format' : config_assembly.eds_ref,
            'path': path
        }

        if connection['input_size'] == 0:
            connection['input_format'] = ""
        if connection['output_size'] == 0:
            connection['output_format'] = ""
        if connection['config_size'] == 0:
            connection['config_format'] = ""

        return connection

    def create_connections(self):
        '''
        Generate connection objects from the device properties according to
        the table below:

        Device type         Exclusive Owner     Listen Only     Input Only
        ------------------------------------------------------------------
        inputs + outputs    Yes                 Yes             Yes
        inputs only         No                  Yes             Yes
        outputs only        Yes                 No              No
        '''
        connection_id = 0

        if self.output_assembly.get_size() > 0:
            connection_id = connection_id + 1
            connection = self.create_connection_dict(
                f'Connection{connection_id}',
                'Point to point',
                '0x84010002','0x44640405',
                '',
                self.rpi_param, self.input_assembly, self.output_assembly, self.config_assembly,
                f'20 04 24 {self.config_assembly.id_hex_str()} 2C {self.output_assembly.id_hex_str()} 2C {self.input_assembly.id_hex_str()}')
            self.connections.append(connection)

        if self.input_assembly.get_size() > 0:
            connection_id = connection_id + 1
            connection = self.create_connection_dict(
                f'Connection{connection_id}',
                'Input Only',
                '0x02030002','0x44640305',
                '',
                self.rpi_param, self.input_assembly, self.input_only_hb_assembly, self.config_assembly,
                f'20 04 24 {self.config_assembly.id_hex_str()} 2C {self.input_only_hb_assembly.id_hex_str()} 2C {self.input_assembly.id_hex_str()}')
            self.connections.append(connection)

        if self.input_assembly.get_size() > 0:
            connection_id = connection_id + 1
            connection = self.create_connection_dict(
                f'Connection{connection_id}',
                'Listen Only',
                '0x01030002','0x44640305',
                '',
                self.rpi_param, self.input_assembly, self.listen_only_hb_assembly, self.config_assembly,
                f'20 04 24 {self.config_assembly.id_hex_str()} 2C {self.listen_only_hb_assembly.id_hex_str()} 2C {self.input_assembly.id_hex_str()}')
            self.connections.append(connection)

    def generate_eds_mapping(self):
        EipAssembly.reset()
        EipParam.reset()

        # Generate input, output and config assemblies matching the
        # modules plugged into slots
        self.input_assembly = EipAssembly('Input Assembly')
        self.output_assembly = EipAssembly('Output Assembly')
        self.config_assembly = EipAssembly('Config Assembly')
        self.input_only_hb_assembly = EipAssembly('Input Only Heartbeat Assembly')
        self.listen_only_hb_assembly = EipAssembly('Listen Only Heartbeat Assembly')

        for slot in self.device.slots:
            module = self.model.get_module(slot.module)
            if len(module.inputs) > 0:
                input_asm  = EipAssembly(f'{slot.name} inputs', module.inputs, param_prefix=slot.name)
                self.input_assembly.subelements.append(input_asm.to_eds_assembly_element())
            if len(module.outputs) > 0:
                output_asm = EipAssembly(f'{slot.name} outputs', module.outputs, param_prefix=slot.name)
                self.output_assembly.subelements.append(output_asm.to_eds_assembly_element())
            if len(module.parameters) > 0:
                config_asm = EipAssembly(f'{slot.name} config', module.parameters, param_prefix=slot.name)
                self.config_assembly.subelements.append(config_asm.to_eds_assembly_element())

        # Create RPI parameter, referenced in the connection objects
        self.rpi_param = EipParam.create_params('RPI', data_type='UINT32',
                                  value_min = str(self.device.ethernetip.min_data_interval),
                                  value_default = str(self.device.ethernetip.default_data_interval))[0]

        self.create_connections()

        EipAssembly.remove_unused_assemblies()

    def generate_eds(self):
        template = self.env.get_template("template.eds")
        self.generate_eds_mapping()
        eds = template.render(
            model=self.model,
            device=self.device,
            params=EipParam.objects,
            assemblies=EipAssembly.objects,
            connections=self.connections)

        _logger.debug(eds)
        return eds



import logging

from enum import Enum
from typing import Optional, List, Union
from pydantic import BaseModel, Field, validator, UUID4, PositiveInt

from .ethercat import *
from .profinet import *
from .ethernetip import *
from .modbus import *

_logger = logging.getLogger(__name__)

HelpText = {
    "Root":{
        "name": "The name of the model. Typically the same as the device name for a single device design.",
        "vendor": "Name of company producing the device. Shall match Profinet and EtherCAT Vendor IDs",
        "adapters": "Enabled fieldbus adapters",
        "devices": "List of devices. Often a single device.",
        "modules": "List of modules. A module is a defines a set of inputs, outputs and more. A module can be instantiated in several slots  ",
        "ethercat": "EtherCAT specific configuration",
        "profinet": "Profinet specific configuration",
        "ethernetip": "EtherNet/IP specific configuration",
        "modbus": "Modbus specific configuration"
    },
    "Device": {
        "name": "The name of the device. Use alphanumeric characters and space to avoid problems.",
        "description": "Short description of functionality. The information will go into device description files and be visible in engineering tools.",
        "hardware_release": "Hardware version string",
        "software_release": "Software version string",
        "serial": "Default serial number as a string. Currently ignored.",
        "loglevel": "Core log level. Currently ignored.",
        "webgui_enable": "Enable device web server. Currently ignored. Webserver is always running. ",
        "slots": "List of slots. Each containing a module.",
        "ethercat": "Device specific EtherCAT configuration",
        "profinet": "Device specific Profinet configuration",
        "ethernetip":"Device specific EtherNet/IP configuration",
        "modbus": "Device specific Modbus configuration",
        "nbr_of_ports": "Number of Ethernet ports on the device"
    },
    "Module": {
        "name": "The name of the module. Will go into device description files and be visible in engineering tools. ",
        "description": "Short description of module functionality. The information will go into device description files and be visible in engineering tools.",
        "inputs": "List of input signals. Input data is generated by device and sent to controller.",
        "outputs": "List of output signals. Output data is generated by controller and received by device",
        "parameters": "List of configuration parameters. Parameters are not part of cyclic process data. Typically set by PLC during startup. ",
        "alarms": "List of supported alarms. Alarms are set by the device to notify the controller on important events",
        "ethercat": "Device specific EtherCAT configuration",
        "profinet": "Module specific Profinet configuration"
    },
    "Slot": {
        "name": "Slot name. U-Phy specific. Not visible in engineering tools",
        "module": "Which module to use in this slot"
    },
    "Alarm": {
        "name": "The alarm name, used for code generation. Not visible in engineering tools",
        "message": "Message shown in PLC applications / SCADA systems" ,
        "error_code": "A number identifying the alarm"
    },
    "BaseSignal": {
        "name": "Signal name. Will be visible in engineering tools.",
        "description": "Short description that will be visible in engineering tools ",
        "datatype": "Signal data type",
        "derived": "Currently not used"
    },
    "Parameter": {
        "default": "Default value",
        "min": "Min value",
        "max": "Max value",
        "persistent": "True if value is stored in non volatile memory",
        "permissions": "Read/Write permissions",
        "profinet": " Profinet parameter index"
    },
    "Signal": {
        "is_array": "True if signal is an array of values",
        "array_length": "Length of signal array",
    }
}

class LogLevel(str, Enum):
    FATAL = 'FATAL'
    ERROR = 'ERROR'
    WARNING = 'WARNING'
    INFO = 'INFO'
    DEBUG = 'DEBUG'

    def __str__(self):
        return self.value


class DataType(str, Enum):
    """Datatypes"""
    UINT32 = 'UINT32'
    UINT16 = 'UINT16'
    UINT8 = 'UINT8'
    INT32 = 'INT32'
    INT16 = 'INT16'
    INT8 = 'INT8'
    REAL32 = 'REAL32'

    def __str__(self):
        return self.value

    @property
    def bitlen(self):
        map = {
            DataType.UINT32: 32,
            DataType.UINT16: 16,
            DataType.UINT8: 8,
            DataType.INT32: 32,
            DataType.INT16: 16,
            DataType.INT8: 8,
            DataType.REAL32: 32,
        }
        return map[self.value]


class Permissions(str, Enum):
    """Read/write permissions"""
    RO = 'RO'
    RW = 'RW'
    WO = 'WO'
    RWpre = 'RWpre'  # Dynamic permission (depends on state)

class Adapter(str, Enum):
    """Available fieldbus adapters"""
    PROFINET = 'Profinet'
    ETHERCAT = 'EtherCAT'
    ETHERNETIP = 'EtherNet/IP'
    MODBUS = 'Modbus'

class ModelElement(BaseModel):
    """Base Class elements with unique id"""
    id: UUID4

class BaseSignal(ModelElement):
    """Base signal defines properties common to signals containing cyclic data and parameters """
    name: str = Field(title="Signal name", description=HelpText['BaseSignal']['name'])
    description: str = Field(title="Description", default="", description=HelpText['BaseSignal']['description'])
    datatype: DataType = Field(title="Signal datatype", description=HelpText['BaseSignal']['datatype'])
    #derived: bool = Field(title="Derived", default=False, description=HelpText['BaseSignal']['derived'])

    @property
    def bitlen(self):
        return self.datatype.bitlen

class Signal(BaseSignal):
    """Signals are cyclic data. They can for instance model a sensor
    or actuator."""
    is_array: Optional[bool] = Field(title="Array", default=False, description=HelpText['Signal']['is_array'])
    array_length: Optional[PositiveInt] = Field(title="Array Length ", description=HelpText['Signal']['array_length'])

    @property
    def bitlen(self):
        if not self.is_array:
            return self.datatype.bitlen
        else:
           return self.datatype.bitlen * self.array_length


class Parameter(BaseSignal):
    """Parameters are acyclic data. They usually model configuration objects."""
    default: str = Field(title="Default value", default="0", description=HelpText['Parameter']['default'])
    min: Optional[str] = Field(title="Min value", description=HelpText['Parameter']['min'])
    max: Optional[str] = Field(title="Max value", description=HelpText['Parameter']['max'])
    #persistent: bool = Field(title="Persistent", default=False, description=HelpText['Parameter']['persistent'])
    permissions: Permissions = Field(title="Permissions", default=Permissions.RW, description=HelpText['Parameter']['permissions'])
    profinet: Optional[ProfinetParameter] = Field(title="Profinet parameter", description=HelpText['Parameter']['profinet'])


class Alarm(ModelElement):
    name: str = Field(title="Alarm name", description=HelpText['Alarm']['name'])
    message: str = Field(title="Alarm message",description=HelpText['Alarm']['message'])
    error_code: str = Field(title="Error code", description=HelpText['Alarm']['error_code'])


class Module(ModelElement):
    """A module is a reusable component that can be instantiated in a slot in a device."""
    name: str = Field(title="Module name", description=HelpText['Module']['name'])
    description: str = Field(title="Description", default="", description=HelpText['Module']['description'])
    inputs: List[Signal] = Field(title="Inputs", default=[], description=HelpText['Module']['inputs'])
    outputs: List[Signal] = Field(title="Outputs", default=[], description=HelpText['Module']['outputs'])
    parameters: List[Parameter] = Field(title="Parameters", default=[], description=HelpText['Module']['parameters'])
    alarms: List[Alarm] = Field(title="Alarm", default=[], description=HelpText['Module']['alarms'])
    ethercat: EtherCATModule = Field(title="EtherCAT module", default=EtherCATModule(), description=HelpText['Module']['ethercat'])
    profinet: ProfinetModule = Field(title="Profinet module", default=ProfinetModule(), description=HelpText['Module']['profinet'])

    def get_signal(self, signals, id):
        return [s for s in signals if s.id == id][0]

    def get_offset(self, signals, s):
        ix = signals.index(s)
        offset = sum([s.bitlen for s in signals[0:ix]])
        return offset

    def get_bitlen(self, signals):
        return sum([s.bitlen for s in signals])

    @property
    def inputs_bitlen(self):
        return self.get_bitlen(self.inputs)

    @property
    def outputs_bitlen(self):
        return self.get_bitlen(self.outputs)

    @property
    def parameters_bitlen(self):
        return self.get_bitlen(self.parameters)


class Slot(BaseModel):
    """A device consists of one or more slots. Each slot contains a module."""
    name: str = Field(title="Slot name", description=HelpText['Slot']['name'])
    module: UUID4 = Field(title="Module", description=HelpText['Slot']['module'])


class Device(ModelElement):
    """A device models a fieldbus device. It consists of one or more
    slots. Each slot contains a module."""
    name: str = Field(title="Device name", description=HelpText['Device']['name'])
    description: str = Field(title="Description", default="", description=HelpText['Device']['description'])
    hardware_release: str = Field(title="Hardware release", default="V0.1.0" , description=HelpText['Device']['hardware_release'])
    software_release: str = Field(title="Software release", default="V0.1.0", description=HelpText['Device']['software_release'])
    serial: Optional[str] = Field(title="Serial number", description=HelpText['Device']['serial'])
    loglevel: LogLevel = Field(title="U-Phy logging level", default=LogLevel.INFO, description=HelpText['Device']['loglevel'])
    webgui_enable: bool = Field(title="Enable webserver", default=True, description=HelpText['Device']['webgui_enable'])
    slots: List[Slot] = Field(title="Slots", description=HelpText['Device']['slots'])
    ethercat: Optional[EtherCATDevice] = Field(title="EtherCAT device", description=HelpText['Device']['ethercat'])
    profinet: Optional[ProfinetDevice] = Field(title="Profinet device", description=HelpText['Device']['profinet'])
    ethernetip: Optional[EthernetipDevice] = Field(title="EtherNet/IP device", description=HelpText['Device']['ethernetip'])
    modbus: Optional[ModbusDevice] = Field(title="Modbus device", description=HelpText['Device']['modbus'])
    nbr_of_ports: Optional[int] = Field(title="Number of ports", default=2, description=HelpText['Device']['nbr_of_ports'])

    def get_used_module_index(self, id, model):
        """Get index in modules list from module id"""
        m = model.get_module(id)
        return self.get_used_modules(model).index(m)

    def get_used_modules(self, model):
        """Get list of modules plugged into slots"""
        used_modules = []
        for s in self.slots:
            if model.get_module(s.module) not in used_modules:
                used_modules.append(model.get_module(s.module))
        return used_modules

    def has_alarms(self, model):
        for m in self.get_used_modules(model):
            if len(m.alarms) > 0:
                return True
        return False

    def get_inputs_frame_size(self, model):
       """Get input frame size in number of bytes"""
       bitlen = 0
       for s in self.slots:
           bitlen += model.get_module(s.module).inputs_bitlen
       return (bitlen + 7) // 8

    def get_outputs_frame_size(self, model):
       """Get output frame size in number of bytes"""
       bitlen = 0
       for s in self.slots:
           bitlen += model.get_module(s.module).outputs_bitlen
       return (bitlen + 7) // 8

class Root(BaseModel):
    """The U-Phy data model is used to define fieldbus devices. A
    device contains one or more slots. Each slot contains one
    module. A module is a reusable component that can be instantiated
    in more than one slot."""
    name: str = Field(title="Model name", description=HelpText['Root']['name'])
    vendor: str = Field(title="Vendor name", description=HelpText['Root']['vendor'])
    adapters: List[Adapter] = Field(
        title="Adapters",
        default=[
            Adapter.PROFINET,
            Adapter.ETHERCAT,
            Adapter.ETHERNETIP,
            Adapter.MODBUS
        ],
        min_items=1,
        uniqueItems=True,
        description=HelpText['Root']['adapters']
    )
    devices: List[Device] = Field(title="Devices", default=[], description=HelpText['Root']['devices'])
    modules: List[Module] = Field(title="Modules", default=[], description=HelpText['Root']['modules'])
    ethercat: Optional[EtherCATRoot] = Field(title="EtherCAT", description=HelpText['Root']['ethercat'])
    profinet: Optional[ProfinetRoot] = Field(title="Profinet",description=HelpText['Root']['profinet'])
    ethernetip: Optional[EthernetipRoot] = Field(title="EtherNet/IP",description=HelpText['Root']['ethernetip'])
    modbus: Optional[ModbusRoot] = Field(title="Modbus",description=HelpText['Root']['modbus'])

    def get_device(self, name):
        return [device for device in self.devices if device.name == name][0]

    def get_module(self, id):
        return [m for m in self.modules if m.id == id][0]

    def get_module_index(self, id):
        """Get index in modules list from module id"""
        m = self.get_module(id)
        return self.modules.index(m)

    @classmethod
    def parse_obj(cls, obj):
        model = super().parse_obj(obj)
        for ix, m in enumerate(model.modules):
            m.ethercat.generate_mapping(m, ix)
            m.profinet.generate_mapping(m, ix)

        return model

    class Config:
        title = 'U-Phy'



import logging
from pathlib import Path
from xmlschema import XMLSchema, etree_tostring
import xml.etree.ElementTree as ET
import json
from . import SCHEMA_DIR
import importlib.resources
from .. import templates
from copy import deepcopy
import os

from ..model import Adapter

xs = XMLSchema(Path(SCHEMA_DIR, 'GSDML/GSDML-DeviceProfile-v2.44.xsd'))

_logger = logging.getLogger(__name__)

pn_type = {
    'UINT32': 'Unsigned32',
    'UINT16': 'Unsigned16',
    'UINT8' : 'Unsigned8',
    'INT32' : 'Integer32',
    'INT16' : 'Integer16',
    'INT8'  : 'Integer8',
    'REAL32': 'Float32'
}

class ProfinetGenerator:
    def __init__(self, model):
        self.model = model

    def import_file(self, file):
        raise Exception('Import not supported ')

    def export_file(self):
        '''Generate GSDML file by applying configuration in a U-Phy model to a template file'''

        if Adapter.PROFINET not in self.model.adapters:
            _logger.debug("Profinet is not enabled")
            return

        if not self.model.profinet:
            _logger.warn('Model does not include a valid profinet configuration')
            return

        self.gsdml_template_name = 'GSDML-V2.44-RT-Labs-U-Phy-Template.xml'
        self.gsdml_file_name = 'GSDML-V2.44-VENDOR-DEVICE-YYYYMMDD.xml'

        ET.register_namespace('', 'http://www.profibus.com/GSDML/2003/11/DeviceProfile')

        gsdml_template_xml = importlib.resources.read_text(templates, self.gsdml_template_name, encoding = "iso-8859-1")

        _logger.info('Generate Profinet GSDML ')
        _logger.info(f"Template: {self.gsdml_template_name}")
        _logger.info(f"Model name: {self.model.name}")

        gsdml_template = xs.to_dict(gsdml_template_xml)

        _logger.info('Apply model data to template')
        gsdml = self.model_to_gsdml(gsdml_template, self.model)

        _logger.debug('Generated gsdml in json format')
        _logger.debug(json.dumps(gsdml, indent=3))

        # Tip: When refactoring code it is useful to temporarily check in
        # a 'gsdml.json' file and compare this with the current output.
        # Enable lines below to write the gsdml json dump to file.
        #file = open('gsdml.json', 'w')
        #file.write(json.dumps(gsdml, indent=3))
        #file.close()

        _logger.info('Convert to xml')
        gsdml_xml = xs.encode(gsdml, encoding = 'iso-8859-1')

        _logger.info('Remove unused attributes to pass gsdml checker')
        all_nodes = gsdml_xml.findall('.//*')
        for node in all_nodes:
            node.attrib.pop('F_IO_StructureDescVersion', None)
            node.attrib.pop('F_IO_StructureDescCRC', None)
            node.attrib.pop('PowerBudgetControlSupported', None)
            node.attrib.pop('BitLength', None)

        _logger.info('Remove RequiredSchemaVersion attribute from ModuleItem nodes')
        module_nodes = gsdml_xml.findall('.//{http://www.profibus.com/GSDML/2003/11/DeviceProfile}ModuleItem')
        for node in module_nodes:
            node.attrib.pop('RequiredSchemaVersion')

        _logger.debug(etree_tostring(gsdml_xml))
        _logger.info(f"Save xml: {str(Path(os.getcwd()) / self.gsdml_file_name)}")
        ET.ElementTree(gsdml_xml).write(self.gsdml_file_name, encoding = 'iso-8859-1')

    def gsdml_to_model(self, gsdml):
        return None

    def model_to_gsdml(self, gsdml_template, up_model):
        return GsdmlGenerator(gsdml_template, up_model).run()

class GsdmlGenerator:
    '''
    Generator for converting a U-Phy model to a gsdml file.

    The generation relies on a template that defines most characteristics.
    GSDML template and and U-Phy device model in dict / json format.
    Limitations:
    - Only one device in model is supported
    - Only characteristics defined by model updated, other characteristics needs to be changed in template, for example
      for a device with other port characteristics.
    - TBD - extend model to cover everything?
    '''
    def __init__(self, gsdml_template, model):
        # Do not update the template but create a working copy
        self.gsdml = deepcopy(gsdml_template)
        self.up_model = model

        # List of TextId objects which is extended during generation
        self.text_list = self.gsdml['ProfileBody']['ApplicationProcess']['ExternalTextList']['PrimaryLanguage']

    def run(self):
        self.gsdml['ProfileBody'] = self.profile_body()
        self.cleanup()
        return self.gsdml

    def profile_body(self):
        up_dev = self.up_model.devices[0]
        profile_body = {
            'DeviceIdentity': {
                '@VendorID': self.up_model.profinet.vendor_id,
                '@DeviceID': self.up_model.profinet.device_id,
                'InfoText': { '@TextId' : self.create_text_id('INFO_DEV', up_dev.description) },
                'VendorName': {'@Value': self.up_model.vendor}
            },
            'DeviceFunction': {
                'Family': {
                    '@MainFamily':  up_dev.profinet.main_family,
                    '@ProductFamily': up_dev.profinet.product_family
                }
            },
            'ApplicationProcess': self.application_process()
        }

        return profile_body

    def application_process(self):
        ap = {}
        ap['DeviceAccessPointList'] = self.device_access_point_list()
        ap['ModuleList'] = self.module_list()
        if self.up_model.devices[0].has_alarms(self.up_model):
            ap['ChannelDiagList'] = self.diagnosis_list()
        ap['ExternalTextList'] = { "PrimaryLanguage": self.text_list }
        return ap

    def device_access_point_list(self):
        dap_list = {
            'DeviceAccessPointItem': []
        }
        for up_dev in self.up_model.devices:
            ix = self.up_model.devices.index(up_dev) + 1
            dap_item = deepcopy(self.gsdml['ProfileBody']['ApplicationProcess']['DeviceAccessPointList']['DeviceAccessPointItem'][0])

            dap_item['@ID'] = f'IDD_{ix}'
            dap_item['@DNS_CompatibleName'] = up_dev.profinet.default_stationname
            dap_item['@ModuleIdentNumber'] = up_dev.profinet.dap_module_id
            dap_item['ModuleInfo'] = self.dap_module_info(up_dev, ix)
            dap_item['UseableModules']['ModuleItemRef'] = self.useable_modules_list(up_dev)
            dap_item['VirtualSubmoduleList']['VirtualSubmoduleItem'][0]["@ID"] = f'IDS_{ix}'
            dap_item['VirtualSubmoduleList']['VirtualSubmoduleItem'][0]["@SubmoduleIdentNumber"] = (
                up_dev.profinet.dap_identity_submodule_id)

            dap_item['VirtualSubmoduleList']['VirtualSubmoduleItem'][0]['ModuleInfo'] = {
                'Name': { "@TextId": f'IDT_DAP_{ix}_NAME' },
                'InfoText': { "@TextId": f'IDT_DAP_{ix}_INFO' }
            }
            dap_item['SystemDefinedSubmoduleList']['InterfaceSubmoduleItem']["@ID"] = f'IDS_{ix}_I'
            dap_item['SystemDefinedSubmoduleList']['InterfaceSubmoduleItem']["@SubmoduleIdentNumber"] = (
            up_dev.profinet.dap_interface_submodule_id)
            dap_item['SystemDefinedSubmoduleList']['PortSubmoduleItem'][0]["@ID"] = f'IDS_{ix}_P1'
            dap_item['SystemDefinedSubmoduleList']['PortSubmoduleItem'][0]["@SubmoduleIdentNumber"] = (
                up_dev.profinet.dap_port_1_submodule_id)
            if up_dev.nbr_of_ports == 2:
                dap_item['SystemDefinedSubmoduleList']['PortSubmoduleItem'][1]["@ID"] = f'IDS_{ix}_P2'
                dap_item['SystemDefinedSubmoduleList']['PortSubmoduleItem'][1]["@SubmoduleIdentNumber"] = (
                up_dev.profinet.dap_port_2_submodule_id)
            else:
                self.remove_text_id('IDT_NAME_PS2')
                del dap_item['SystemDefinedSubmoduleList']['PortSubmoduleItem'][1]

            dap_list['DeviceAccessPointItem'].append(dap_item)
        return dap_list

    def useable_modules_list(self, up_dev):
        ul = []
        pn_slot = 1
        plug_type = '@FixedInSlots'

        for slot in up_dev.slots:
            added = False
            module = self.up_model.get_module(slot.module)
            useable_module = self.create_useable_module(module, pn_slot, plug_type)
            for um in ul:
                if um['@ModuleItemTarget'] == useable_module['@ModuleItemTarget']:
                    um[plug_type] += f" {str(pn_slot)}"
                    added = True

            if not added:
                ul.append(useable_module)

            pn_slot = pn_slot + 1
        return ul

    def dap_module_info(self, up_dev, ix):
        module_info = {
            'Name': {'@TextId': self.create_text_id(f'DAP_{ix}_NAME', up_dev.name)},
            'InfoText': {'@TextId': self.create_text_id(f'DAP_{ix}_INFO', up_dev.description)},
            'VendorName': {'@Value': self.up_model.vendor},
            'OrderNumber': {'@Value': up_dev.profinet.order_id},
            'HardwareRelease': {'@Value': up_dev.hardware_release},
            'SoftwareRelease': {'@Value': up_dev.software_release}
        }
        return module_info

    def create_text_id(self, id, value):
        text_item = {
            '@TextId': f"IDT_{id.strip().replace(' ', '_')}",
            '@Value' : value
        }
        self.text_list['Text'].append(text_item)
        return text_item['@TextId']

    def create_useable_module(self, up_module, pn_slot, plug_type):
        module_item_ref = {}
        module_item_ref['@ModuleItemTarget'] = f"IDM_{up_module.profinet.module_id}"
        module_item_ref[plug_type] = str(pn_slot)
        return module_item_ref

    def module_list(self):
        mod_list = {
            'ModuleItem': []
        }
        for module in self.up_model.modules:
            mod_list['ModuleItem'].append(self.create_module_item(module))

        return mod_list

    def create_module_item(self, up_module):
        module_item = {
            '@ID': f"IDM_{up_module.profinet.module_id}",
            '@ModuleIdentNumber': up_module.profinet.module_id,
            'ModuleInfo': {
                'Name': {
                    '@TextId': self.create_text_id(f"MOD_NAME_{up_module.profinet.module_id}", up_module.name)},
                'InfoText': {
                    '@TextId': self.create_text_id(f"MOD_INFO_{up_module.profinet.module_id}", up_module.description)
                }
            },
            'VirtualSubmoduleList': {
                'VirtualSubmoduleItem': [
                    self.create_virtual_sub_module_item(up_module)
                ]
            }
        }

        return module_item

    def diagnosis_list(self):
        diag_list = {
            'ChannelDiagItem': []
        }
        for module in self.up_model.modules:
            for alarm in module.alarms:
                diag_list['ChannelDiagItem'].append(self.create_diag_item(alarm))
        return diag_list

    def create_diag_item(self, alarm):
        diag_item = {
            '@ErrorType': int(alarm.error_code, 0),
            '@MaintenanceAlarmState':"D;MR;MD",
            'Name': { '@TextId': self.create_text_id(f"DIAG_{str(int(alarm.error_code, 0))}", alarm.message) }
        }
        return diag_item

    def create_virtual_sub_module_item(self, up_module):
        item = {
            '@ID': f"IDSM_{up_module.profinet.submodule_id}",
            '@SubmoduleIdentNumber': up_module.profinet.submodule_id,
            '@MayIssueProcessAlarm': False,
            '@API': 0,
            'IOData': self.create_io_data(up_module),
        }

        if len(up_module.parameters) > 0:
            item['RecordDataList'] = self.create_record_data_list(up_module)

        item['ModuleInfo'] = {
            'Name': {'@TextId': self.create_text_id(f"SUBMOD_NAME_{up_module.profinet.submodule_id}", up_module.name) },
            'InfoText': { '@TextId': self.create_text_id(f"SUBMOD_INFO_{up_module.profinet.submodule_id}", up_module.description) }
        }

        return item

    def create_io_data(self, up_module):
        io_data = {}
        if len(up_module.inputs) > 0:
            io_data['Input'] = {}
            io_data['Input']['DataItem'] = []
            for signal in up_module.inputs:
                if not signal.is_array:
                    data_item = {
                        '@DataType': pn_type[signal.datatype],
                        '@TextId': self.create_text_id(f"{up_module.name}_IN_{signal.name}", signal.name),
                        '@UseAsBits': False,
                        '@Subordinate': False
                    }
                    io_data['Input']['DataItem'].append(data_item)
                else:
                    for i in range(0, signal.array_length):
                        data_item = {
                            '@DataType': pn_type[signal.datatype],
                            '@TextId': self.create_text_id(f"{up_module.name}_IN_{signal.name}[{str(i)}]", f"{signal.name} [{str(i)}]"),
                            '@UseAsBits': False,
                            '@Subordinate': False
                        }
                        io_data['Input']['DataItem'].append(data_item)

        if len(up_module.outputs) > 0:
            io_data['Output'] = {}
            io_data['Output']['DataItem'] = []

            for signal in up_module.outputs:
                if not signal.is_array:
                    data_item = {
                        '@DataType': pn_type[signal.datatype],
                        '@TextId': self.create_text_id(f"{up_module.name}_OUT_{signal.name}", signal.name),
                        '@UseAsBits': False,
                        '@Subordinate': False
                    }
                    io_data['Output']['DataItem'].append(data_item)
                else:
                    for i in range(0, signal.array_length):
                        data_item = {
                            '@DataType': pn_type[signal.datatype],
                            '@TextId': self.create_text_id(f"{up_module.name}_OUT_{signal.name}[{str(i)}]", f"{signal.name} [{str(i)}]"),
                            '@UseAsBits': False,
                            '@Subordinate': False
                        }
                        io_data['Output']['DataItem'].append(data_item)


        return io_data

    def create_record_data_list(self, up_module):
        rec_list = []
        for par in up_module.parameters:
            rec_item = {
                '@Index': int(par.profinet.index, 0),
                '@Length': par.bitlen // 8,
                'Name': {
                    '@TextId': self.create_text_id(f"REC_NAME_{par.name}", par.name)
                },
                'Ref': {
                    '@DataType': pn_type[par.datatype],
                    '@DefaultValue': par.default,
                    '@ByteOffset': 0 ,
                    '@TextId': self.create_text_id(f"REC_INFO_{par.name}", par.description)
                }
            }
            if par.min is not None and par.max is not None:
                rec_item['Ref']['@AllowedValues'] = f"{par.min}..{par.max}"
            rec_list.append(rec_item)

        return {'ParameterRecordDataItem': rec_list }


    def cleanup(self):
        ''' Remove unused TextId from original template '''
        self. remove_text_id('IDT_DUMMY')

    def remove_text_id(self, text_id):
        for text in self.text_list['Text']:
            if text['@TextId'] == text_id:
                self.text_list['Text'].remove(text)



import argparse
import logging
import sys
import os
import traceback
from argparse import RawTextHelpFormatter

from pathlib import Path
from json.decoder import JSONDecodeError
from pydantic import ValidationError

from upgen import __version__
from upgen.model import *
from upgen.generators import EtherCATGenerator
from upgen.generators import ProfinetGenerator
from upgen.generators import EthernetipGenerator
from upgen.generators import CCodeGenerator
from upgen.generators import SIIGenerator

_logger = logging.getLogger(__name__)

GENERATORS = {
    'EtherCAT': EtherCATGenerator,
    'Profinet': ProfinetGenerator,
    "EtherNetIP":EthernetipGenerator,
        'Code': CCodeGenerator,
         'SII': SIIGenerator,

}


def get_generator_classes(name):
    if name == 'All':
        return [GENERATORS[g] for g in GENERATORS]
    try:
        return [GENERATORS[name]]
    except KeyError:
        raise ValueError(f"--generator name must be in {tuple(GENERATORS)} or \'All\')")


def parse_args(args):
    parser = argparse.ArgumentParser(description=
    """
    Tool for generation of device artifacts from a U-Phy device model
    file, including:
    - C-code with the device specific data model
    - EtherCAT ESI file
    - EtherCAT SII EEPROM file
    - Profinet GSDML file

    To evaluate the device functionality, copy generated model.c/h and
    eeprom.bin to the \'generated\' folder of the U-Phy sample application
    and rebuild.

    Example usage:
        upgen export my_device_model.json
    """,
    formatter_class=RawTextHelpFormatter)
    parser.add_argument(
        "--version",
        action="version",
        version=f"upgen {__version__}",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="loglevel",
        help="set loglevel to INFO",
        action="store_const",
        const=logging.INFO,
    )
    parser.add_argument(
        "-vv",
        "--very-verbose",
        dest="loglevel",
        help="set loglevel to DEBUG",
        action="store_const",
        const=logging.DEBUG,
    )
    parser.add_argument("-d", "--destination", help="destination folder for generated model", default=".")

    parser.add_argument('--generator', help="the generator to use. Option value "
                        "can be one of  {} or \'All\'. Default value is \'All\'".format(tuple(GENERATORS)), default="All")

    parser.add_argument('-n', '--device_name', help="the name of the device to generate. "
                        "Only used when generating C code or SII eeprom for multi device models.")

    # Import device description file or
    # Export source code and device description files from json model
    # TODO - for export the generators for fieldbuses defined by the
    # model should be run. All outputs placed in the destination folder.
    parser.add_argument('action', choices=('import', 'export', 'schema'))
    parser.add_argument("file", help="Device model in JSON format")

    return parser.parse_args(args)


def setup_logging(loglevel):
    logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
    if loglevel is None:
        loglevel = logging.WARNING
    logging.basicConfig(
        level=loglevel, stream=sys.stdout, format=logformat, datefmt="%Y-%m-%d %H:%M:%S"
    )


def main(args):
    args = parse_args(args)
    setup_logging(args.loglevel)

    if args.action == 'import':
        raise Exception( "import not supported")
        generator_class = get_generator_class(args.generator)
        generator = generator_class(None)
        generator.import_file(args.file)
        print(generator.model)

    elif args.action == 'export':
        try:
            model = Root.parse_file(Path(args.file))
        except (JSONDecodeError, ValidationError) as e:
            print(e)
            sys.exit(1)

        os.makedirs(args.destination, exist_ok=True)
        os.chdir(args.destination)
        generator_classes = get_generator_classes(args.generator)
        for gen_cls in generator_classes:
            try:
                generator = gen_cls(model)
                if (gen_cls in [CCodeGenerator, SIIGenerator, EthernetipGenerator]):
                    generator.select_device(args.device_name)
                generator.export_file()
            except Exception as e:
                print(
                    f"Error while running {gen_cls.__name__}: \"{str(e)}\" Continue generating rest of artifacts.")
                _logger.warning(''.join(traceback.TracebackException.from_exception(e).format()))

    elif args.action == 'schema':
        print(Root.schema_json(indent=2))

    _logger.info("Script ends here")


def run():
    main(sys.argv[1:])


if __name__ == "__main__":
    run()



from io import BufferedIOBase
import logging
from pathlib import Path
import os

from ctypes import LittleEndianStructure, sizeof, c_ubyte, c_ushort, c_uint
from crc8 import crc8

from ..model import Adapter

_logger = logging.getLogger(__name__)


def str2hex(str):
    try:
        return int(str.replace('#', '0'), base=0)
    except (ValueError, AttributeError):
        return 0


def eeprom_size(nbytes):
    bits = nbytes * 8
    kbits = bits // 1024
    return kbits - 1


class Category:
    DC = 60
    RXPDO = 51
    TXPDO = 50
    SYNCM = 41
    FMMU = 40
    GENERAL = 30
    STRINGS = 10


class SIIElement(LittleEndianStructure):

    def write(self, f):
        f.write(self)


class SIIHeader(SIIElement):
    _fields_ = [
        ("category", c_ushort),
        ("wordSize", c_ushort)
    ]

    def __init__(self, category, size):
        self.category = category
        self.is_odd = size & 0x01
        if self.is_odd:
            # Pad data to even number of bytes
            size += 1
        self.wordSize = size // 2


class SIIPdi(SIIElement):
    _fields_ = [
        ("pdiControl", c_ushort),
        ("pdiConfiguration", c_ushort),
        ("syncImpulseLen", c_ushort),
        ("pdiConfiguration2", c_ushort),
        ("stationAlias", c_ushort),
        ("asicConfig", c_ushort),
        ("reserved1", c_ushort),
        ("checksum", c_ushort),
        ("vendorID", c_uint),
        ("productCode", c_uint),
        ("revisionNumber", c_uint),
        ("serialNumber", c_uint),
        ("reserved2", c_ubyte * 8),
        ("bootRecvMboxOffset", c_ushort),
        ("bootRecvMboxSize", c_ushort),
        ("bootSendMboxOffset", c_ushort),
        ("bootSendMboxSize", c_ushort),
        ("standardRecvMboxOffset", c_ushort),
        ("standardRecvMboxSize", c_ushort),
        ("standardSendMboxOffset", c_ushort),
        ("standardSendMboxSize", c_ushort),
        ("mboxProtocol", c_ushort),
        ("reserved3", c_ubyte * 66),
        ("eepromSize", c_ushort),
        ("version", c_ushort),
    ]

    def write(self, f):
        # Calculate CRC8 checksum of first 7 words
        hash = crc8(initial_start=0xFF)
        b = bytearray(self)
        hash.update(b[0:14])
        self.checksum = ord(hash.digest())
        f.write(self)


class SIICategoryGeneral(SIIElement):
    _fields_ = [
        ("groupIdx", c_ubyte),
        ("imgIdx", c_ubyte),
        ("orderIdx", c_ubyte),
        ("nameIdx", c_ubyte),
        ("reserved1", c_ubyte),
        ("coeDetails", c_ubyte),
        ("foeDetails", c_ubyte),
        ("eoeDetails", c_ubyte),
        ("reserved2", c_ubyte * 3),
        ("flags", c_ubyte),
        ("currentOnEbus", c_ushort),
        ("reserved3", c_ubyte * 2),
        ("physicalPort", c_ushort),
        ("reserved4", c_ubyte * 14),
    ]

    def write(self, f):
        header = SIIHeader(Category.GENERAL, sizeof(self))
        f.write(header)
        f.write(self)


class SIIFmmu(SIIElement):
    _fields_ = [
        ("fmmu", c_ubyte),
    ]


class SIICategoryFmmu:

    def __init__(self, fmmus):
        self.fmmus = fmmus
        self.size = sum(map(sizeof, fmmus))

    def write(self, f):
        header = SIIHeader(Category.FMMU, self.size)
        f.write(header)
        for fmmu in self.fmmus:
            f.write(fmmu)
        if header.is_odd:
            f.write(c_ubyte())


class SIISyncM(SIIElement):
    _fields_ = [
        ("start", c_ushort),
        ("length", c_ushort),
        ("control", c_ubyte),
        ("status", c_ubyte),
        ("enable", c_ubyte),
        ("type", c_ubyte),
    ]


class SIICategorySyncM:

    def __init__(self, syncms):
        self.syncms = syncms
        self.size = sum(map(sizeof, syncms))

    def write(self, f):
        header = SIIHeader(Category.SYNCM, self.size)
        f.write(header)
        for syncm in self.syncms:
            f.write(syncm)


def sii_string(string):
    '''Factory function for strings, as the length depends on the string'''
    class SIIString(SIIElement):
        _fields_ = [
            ("length", c_ubyte),
            ("string", c_ubyte * len(string)),
        ]

        def __init__(self, string):
            self.length = len(string)
            self.string = (c_ubyte*len(string))(*bytearray(string, "utf-8"))

    return SIIString(string)


class SIICategoryString:

    def __init__(self, strings):
        self.strings = strings
        self.size = sum(map(sizeof, strings))

    def write(self, f):
        header = SIIHeader(Category.STRINGS, self.size + 1)
        f.write(header)
        f.write(c_ubyte(len(self.strings)))
        for string in self.strings:
            f.write(string)
        if header.is_odd:
            f.write(c_ubyte())


class SIIGenerator:

    def __init__(self, model):
        self.model = model
        self.device = self.model.devices[0]

    def select_device(self, device_name):
        if device_name == None:
            self.device = self.model.devices[0]
        else:
            self.device = self.model.get_device(device_name)

        if self.device == None:
            raise Exception(f'{device_name} not found in model')

    def export_file(self):
        if Adapter.ETHERCAT not in self.model.adapters:
            _logger.debug("EtherCAT is not enabled")
            return

        if not self.model.ethercat:
            _logger.warn('Model does not include a valid EtherCAT configuration')
            return

        _logger.info("Exporting EtherCAT SII eeprom contents")

        filename = str(Path(os.getcwd()) / "eeprom.bin")
        with open(filename, "wb") as f:
            self.export_data(f)

    def export_data(self, f):
        ecat = self.model.ethercat
        device = self.device

        pdi = SIIPdi()
        pdi.pdiControl = ecat._pdi_control
        pdi.pdiConfiguration = ecat._pdi_configuration
        pdi.asicConfig = ecat._asic_config
        pdi.vendorID = str2hex(ecat.vendor_id)
        pdi.productCode = str2hex(device.ethercat.product_code)
        pdi.revisionNumber = str2hex(device.ethercat.revision)
        pdi.serialNumber = str2hex(device.serial)
        pdi.bootRecvMboxOffset = ecat._boot_recv_mbox.offset
        pdi.bootRecvMboxSize = ecat._boot_recv_mbox.size
        pdi.bootSendMboxOffset = ecat._boot_send_mbox.offset
        pdi.bootSendMboxSize = ecat._boot_send_mbox.size
        pdi.standardRecvMboxOffset = ecat._standard_recv_mbox.offset
        pdi.standardRecvMboxSize = ecat._standard_recv_mbox.size
        pdi.standardSendMboxOffset = ecat._standard_send_mbox.offset
        pdi.standardSendMboxSize = ecat._standard_send_mbox.size
        pdi.mboxProtocol = ecat._mbox_control
        pdi.eepromSize = eeprom_size(ecat._eeprom_size)
        pdi.version = 1

        strings = SIICategoryString([
            # The string addition order is important, since SOEM will
            # choose the first string as the name
            sii_string(device.name),                 # Name
            sii_string(self.model.ethercat.group)    # Group
        ])

        general = SIICategoryGeneral()

        # Set index into strings array (index starts at 1)
        general.groupIdx = 2  # model.ethercat.group
        general.orderIdx = 1  # device.name
        general.nameIdx = 1  # device.name

        general.reserved1 = 5  # Port 0 & 1: MII
        general.coeDetails = ecat._coe
        general.foeDetails = ecat._foe

        fmmu = SIICategoryFmmu([
            SIIFmmu(f.value) for f in ecat._fmmu
        ])

        sm = SIICategorySyncM([
            SIISyncM(
                sm.address,
                sm.length,
                sm.control,
                sm.status,
                sm.enable,
                sm.type.value,
            ) for sm in ecat._sm
        ])

        pdi.write(f)
        strings.write(f)
        general.write(f)
        fmmu.write(f)
        sm.write(f)

        # Fill remainder with 0xFF
        remain = ecat._eeprom_size - f.tell()
        for _ in range(remain):
            f.write(c_ubyte(0xFF))



from enum import Enum
from typing import Optional, List
from pydantic import BaseModel, Field, validator

HelpText = {
    "EthernetipRoot": {
        "vendor_id": "Vendor id assigned by ODVA organization"
    },
    "EthernetipDevice": {
        "revision": "Device revision in format major.minor . Increment major revision if device’s logical interface changes. Increment minor revision if behavior on network changes",
        "product_code": "Product code, defined by vendor",
        "device_type": "Type of device. If uncertain, use 43 (Generic Device)",
        "device_type_str": "Type of device. If uncertain, use Generic Device",
        "home_url": "URL for company or product",
        "create_date": "EDS file creation date. If string is empty current date will be used.",
        "create_time": "EDS file creation time. If string is empty current time will be used.",
        "modification_date" : "EDS file modification date. If string is empty current date will be used.",
        "modification_time": "EDS file modification time. If string is empty current time will be used.",
        "min_data_interval": "Fastest datarate supported by device, time in micro seconds",
        "default_data_interval": "Default datarate for device, time in micro seconds"
    }
}

class EthernetipDevice(BaseModel):
    revision: str = Field(title="Revision", default="1.1", description=HelpText['EthernetipDevice']['revision'])
    product_code: str = Field(title="Product Code", description=HelpText['EthernetipDevice']['product_code'])
    device_type: str = Field(title="Device Type", default="43", description=HelpText['EthernetipDevice']['device_type'])
    device_type_str: str = Field(title="Device Type String", default="Generic Device", description=HelpText['EthernetipDevice']['device_type_str'])
    home_url:str = Field(title="Home URL", default="https://rt-labs.com/u-phy/", description=HelpText['EthernetipDevice']['home_url'])
    create_date:Optional[str] = Field(title="EDS File Creation Date", description=HelpText['EthernetipDevice']['create_date'])
    create_time:Optional[str] = Field(title="EDS File Creation Time", description=HelpText['EthernetipDevice']['create_time'])
    modification_date:Optional[str] = Field(title="EDS File Modification Date", description=HelpText['EthernetipDevice']['modification_date'])
    modification_time:Optional[str] = Field(title="EDS File Modification Time", description=HelpText['EthernetipDevice']['modification_time'])
    min_data_interval: str = Field(title="RPI Min - Minimal cyclic data period (in microseconds)", default="4000", description=HelpText['EthernetipDevice']['min_data_interval'])
    default_data_interval: str = Field(title="RPI Default - Default cyclic data period (in microseconds)", default="10000",description=HelpText['EthernetipDevice']['default_data_interval'])

class EthernetipRoot(BaseModel):
    vendor_id: str = Field(title="Vendor Id", default="1772", description=HelpText['EthernetipRoot']['vendor_id'])



from enum import Enum
from typing import Optional, List
from pydantic import BaseModel, Field, validator


HelpText = {
    "ProfinetParameter": {
        "index": "Parameter index used by PLC to address the parameter"
    },
    "ProfinetModule": {
        "module_id": "Module Id. Used by PLC to address the module. ",
        "submodule_id": "Submodule Id. U-Phy implements one submodule per module"
    },
    "ProfinetDevice": {
        "dap_module_id": "Device Access Point module identity number",
        "dap_identity_submodule_id": "No help text available",
        "dap_interface_submodule_id": "No help text available",
        "dap_port_1_submodule_id": "No help text available",
        "dap_port_2_submodule_id": "No help text available",
        "profile_id": "No help text available",
        "profile_specific_type": "No help text available",
        "min_device_interval": "Minimum time interval for sending cyclic IO data. Supported values [32,64,128,256]. Profinet time base is 1/32 ms and 32 gives a send interval of 1ms.",
        "default_stationname": "Name used detect and configure the device using Profinet engineering tools (Discovery and Configuration Protocol)",
        "order_id": "Device order information",
        "hw_revision": "No help text available",
        "sw_revision_prefix": "No help text available",
        "sw_revision_functional_enhancement": "No help text available",
        "sw_revision_bug_fix": "No help text available",
        "sw_revision_internal_change": "No help text available",
        "revision_counter": "No help text available",
        "main_family": "Functional class. For example I/O, Sensor or Gateway",
        "product_family": "Vendor defined product family. Used by engineering tool to generate a catalog hiararchy."
    },
    "ProfinetRoot": {
        "vendor_id": "Vendor identity assigned by PI organization",
        "device_id": "Device identity. Common to all devices in the model. The DAP Module ID is used to identify devices within model."
    }
}


class MainFamilyType(str, Enum):
    """Type of functionaliy. Used by engineering tool to group devices"""
    GENERAL = 'General'
    DRIVES = 'Drives'
    SWITCHING_DEVICES = 'Switching Devices'
    IO = 'I/O'
    VALVES = 'Valves'
    CONTROLLERS = 'Controllers'
    HMI = 'HMI'
    ENCODERS = 'Encoders'
    NC_RC = 'NC/RC'
    GATEWAY = 'Gateway'
    PLCS = 'PLCs'
    IDENT_SYSTEMS = 'Ident Systems'
    PA_PROFILES = 'PA Profiles'
    NETWORK_COMPONENTS = 'Network Components'
    SENSORS = 'Sensors'

    def __str__(self):
        return self.value

def hex_uint32(value):
    return f'0x{int(value,0):08x}'

def hex_uint16(value):
    return f'0x{int(value,0):04x}'

class ProfinetParameter(BaseModel):
    """Profinet specific parameter configuration"""
    index: str = Field(title="Parameter index", default="")

class ProfinetModule(BaseModel):
    """Profinet specific module configuration"""
    module_id: Optional[str] = Field(title="Module ID", default="")
    submodule_id: Optional[str] = Field(title="Submodule ID", default="")

    @validator('module_id', 'submodule_id')
    def hex32(cls, v):
        if not v:
            return ""
        return hex_uint32(v)

    def generate_mapping(self, module, module_ix):
        module_id = (module_ix + 1) * 0x100

        if not self.module_id:
            self.module_id = f"{module_id:#x}"

        if not self.submodule_id:
            self.submodule_id = f"{module_id + 1:#x}"

        for ix, parameter in enumerate(module.parameters):
            if not parameter.profinet:
                parameter.profinet = ProfinetParameter(
                    index = f"{ix + 0x100:#x}"
                )

class ProfinetDevice(BaseModel):
    """Profinet specific device configuration"""
    dap_module_id: str = Field(title="DAP Module ID", default="1", description=HelpText['ProfinetDevice']['dap_module_id'])
    dap_identity_submodule_id: str = Field(title="DAP Identity Submodule ID", default="0x00000001", description=HelpText['ProfinetDevice']['dap_identity_submodule_id'])
    dap_interface_submodule_id: str = Field(title="DAP Interface Submodule ID", default="0x00008000", description=HelpText['ProfinetDevice']['dap_interface_submodule_id'])
    dap_port_1_submodule_id: str = Field(title="DAP Port 1 Submodule ID", default="0x00008001", description=HelpText['ProfinetDevice']['dap_port_1_submodule_id'])
    dap_port_2_submodule_id: str = Field(title="DAP Port 2 Submodule ID", default="0x00008002", description=HelpText['ProfinetDevice']['dap_port_2_submodule_id'])

    @validator('dap_module_id', 'dap_identity_submodule_id', 'dap_interface_submodule_id', 'dap_port_2_submodule_id' )
    def hex32(cls, v):
        return hex_uint32(v)

    @validator('profile_id','profile_specific_type', )
    def hex16(cls, v):
        return hex_uint16(v)

    profile_id: str = Field(title="Profile ID", default="0", description=HelpText['ProfinetDevice']['profile_id'])
    profile_specific_type: str = Field(title="Profile specific type", default="0", description=HelpText['ProfinetDevice']['profile_specific_type'])
    min_device_interval: str = Field(title="Min device interval", default="32", description=HelpText['ProfinetDevice']['min_device_interval'])
    default_stationname: str = Field(title="Default station name", description=HelpText['ProfinetDevice']['default_stationname'])
    order_id: str = Field(title="Order ID", description=HelpText['ProfinetDevice']['order_id'])
    hw_revision: str = Field(title="Hardware revision", default="0", description=HelpText['ProfinetDevice']['hw_revision'])
    sw_revision_prefix: str = Field(title="Software revision prefix", default="V", description=HelpText['ProfinetDevice']['sw_revision_prefix'])
    sw_revision_functional_enhancement: str = Field(title="Software revision functional enhancement", default="0", description=HelpText['ProfinetDevice']['sw_revision_functional_enhancement'])
    sw_revision_bug_fix: str = Field(title="Software revision bug fix", default="0", description=HelpText['ProfinetDevice']['sw_revision_bug_fix'])
    sw_revision_internal_change: str = Field(title="Software revision internal change", default="0", description=HelpText['ProfinetDevice']['sw_revision_internal_change'])
    revision_counter: str = Field(title="Revision counter", default="0", description=HelpText['ProfinetDevice']['revision_counter'])
    main_family: MainFamilyType = Field(title="Main family", default=MainFamilyType.IO, description=HelpText['ProfinetDevice']['main_family'])
    product_family: str = Field(title="Product family", description=HelpText['ProfinetDevice']['product_family'])

class ProfinetRoot(BaseModel):
    """Profinet config"""
    vendor_id: str = Field(title="Vendor ID", description=HelpText['ProfinetRoot']['vendor_id'])
    device_id: str = Field(title="Device Family ID", description=HelpText['ProfinetRoot']['device_id'])

    @validator('vendor_id', 'device_id')
    def hex16(cls, v):
        return hex_uint16(v)



from enum import Enum
from typing import Optional, List
from pydantic import BaseModel, Field, validator

HelpText = {
    "ModbusDevice":
    {
        "port": "TCP port.",
    },
}

def hex_uint16(value):
    return f'0x{int(value,0):04x}'
class ModbusDevice(BaseModel):
    port: str = Field(title = "Port", default="502", description=HelpText["ModbusDevice"]["port"])

    @validator('port')
    def hex16(cls, v):
        return hex_uint16(v)

class ModbusRoot(BaseModel):
    pass





# import os
from pathlib import Path, PurePath

SCHEMA_DIR = Path(PurePath(__file__).parent, "schema")

from .ethercat import EtherCATGenerator
from .sii import SIIGenerator
from .profinet import ProfinetGenerator
from .ethernetip import EthernetipGenerator
from .c_code import CCodeGenerator



import logging

from pathlib import Path
from xmlschema import XMLSchema, etree_tostring
import xml.etree.ElementTree as ET
from . import SCHEMA_DIR
import os


from ..model import Adapter

xs = XMLSchema(Path(SCHEMA_DIR, 'ESI/EtherCATInfo.xsd'))

_logger = logging.getLogger(__name__)

slot_pdo_increment = 16
slot_index_increment = 0x800

ecat_type = {
    'UINT32': 'UDINT',
    'UINT16': 'UINT',
    'UINT8': 'USINT',
    'INT32': 'DINT',
    'INT16': 'INT',
    'INT8': 'SINT',
    'REAL32':'REAL',
}

ecat_bitlen = {
    'UDINT':32,
    'UINT':16,
    'USINT':8,
    'DINT':32,
    'INT':16,
    'SINT':8,
    'REAL':32,
    'USINT':8
}


default_types = [
    {
        'Name': 'ULINT',
        'BitSize': 64
    },
    {
        'Name': 'UDINT',
        'BitSize': 32
    },
    {
        'Name': 'UINT',
        'BitSize': 16
    },
    {
        'Name': 'USINT',
        'BitSize': 8
    },
    {
        'Name': 'DINT',
        'BitSize': 32
    },
    {
        'Name': 'INT',
        'BitSize': 16
    },
    {
        'Name': 'SINT',
        'BitSize': 8
    },
    {
        'Name': 'REAL',
        'BitSize': 32
    },
    {
        'Name': 'LREAL',
        'BitSize': 64
    },
    {
        'Name': 'BOOL',
        'BitSize': 1
    },
    {
        'Name': 'BYTE',
        'BitSize': 8
    },
    {
        'Name': 'ARRAY [0..15] OF BYTE',
        'BaseType':'BYTE',
        'BitSize': 128,
        'ArrayInfo':{
            'LBound':0,
            'Elements':16
        }
    }
]


def hexify(s):
    return s.replace('0x', '#x')

def dehexify(s):
    return s.replace("#x", "0x")

class EtherCATGenerator:
    def __init__(self, model):
        self.model = model
        self.esi_file_name = f"{self.model.name}.xml"
        self.data_types = {}
        for t in default_types:
            self.add_data_type(t)

    def import_file(self, file):
        _logger.info('Importing model')
        esi = xs.decode(file)
        _logger.debug(esi)
        self.model = self.esi_to_model(esi)

    def export_file(self):
        if Adapter.ETHERCAT not in self.model.adapters:
            _logger.debug("EtherCAT is not enabled")
            return

        if not self.model.ethercat:
            _logger.warn('Model does not include a valid EtherCAT configuration')
            return

        _logger.info('Exporting EtherCAT ESI file')

        esi = self.model_to_esi()
        xml = xs.encode(esi)
        _logger.debug(etree_tostring(xml))
        _logger.info(f"Save xml: {str(Path(os.getcwd()) / self.esi_file_name)}")
        ET.ElementTree(xml).write(self.esi_file_name, encoding='iso-8859-1')

    def esi_to_model(self, esi):
        return None

    def model_to_esi(self):
        config = self.model.ethercat
        esi = {
            'Vendor': {
                'Id': config.vendor_id,
                'Name': self.model.vendor,
            },
            'Descriptions': {
                'Groups': self.model_to_groups(),
                'Devices': self.model_to_devices(),
                'Modules': self.model_to_modules(),
            },
        }
        return esi

    def model_to_groups(self):
        return {
            'Group': [
                {
                    'Type': self.model.ethercat.group,
                    'Name': self.model.ethercat.group
                }
            ]
        }

    def sm_obj(self, cfg):
        sm = {
            '@ControlByte': hexify(f'0x{cfg.control:02x}'),
            '@DefaultSize': str(cfg.length),
            '@Enable': str(cfg.enable),
            '@StartAddress': hexify(f'0x{cfg.address:04x}'),
            '$': cfg.type.name
        }
        return sm

    def model_to_devices(self):
        devices = []
        for d in self.model.devices:
            meta = d.ethercat
            device = {
                '@Physics': 'YY',
                'Type': {
                    '@ProductCode': meta.product_code,
                    '@RevisionNo': meta.revision,
                    '$': d.name
                },
                'Name': d.name,
                'Info': {
                    'StateMachine': {
                        'Timeout': {
                            'PreopTimeout': 1000,
                            'SafeopOpTimeout': 1000,
                            'BackToInitTimeout': 1000,
                            'BackToSafeopTimeout': 200,
                        },
                    },
                },
                'GroupType': self.model.ethercat.group,
                'Profile': {
                    'ProfileNo': int(dehexify(meta.profile), base=0),
                    'AddInfo': 0,
                    'Dictionary': self.model_to_directory(d)
                },
                'Fmmu': [f.name for f in self.model.ethercat._fmmu],
                'Sm': [self.sm_obj(sm) for sm in self.model.ethercat._sm],
                'Mailbox': {
                    '@DataLinkLayer': True,
                    'CoE': {
                        '@CompleteAccess': True,
                        '@PdoUpload': False,
                        '@SdoInfo': True,
                        '@DiagHistory': True
                    },
                    'FoE': {},
                },
                'Slots': {
                    '@SlotPdoIncrement': str(slot_pdo_increment),
                    '@SlotIndexIncrement': hexify(f'0x{slot_index_increment:x}'),
                    'Slot': self.model_to_slots(d)
                }
            }

            if d.has_alarms(self.model):
                device['Profile']['DiagMessages'] = self.model_diag_messages(d)

            devices.append(device)

        return {'Device': devices}

    def model_to_slots(self, device):
        slots = []
        for s in device.slots:
            slot = {
                '@MinInstances': '1',
                '@MaxInstances': '1',
                'Name': s.name,
                'ModuleIdent': {
                    '@Default': '1',
                    '$': str(self.model.get_module_index(s.module) + 1),
                }
            }
            slots.append(slot)

        return slots

    def model_to_modules(self):
        modules = []
        for (ix, m) in enumerate(self.model.modules):
            module = {
                'Type': {
                    '@ModuleIdent': str(ix + 1),
                    '$': m.name
                },
                'Name': m.name,
            }
            rxpdos = self.model_to_module_rxpdo(m)
            if len(rxpdos) > 0:
                module['RxPdo'] = rxpdos
            txpdos = self.model_to_module_txpdo(m)
            if len(txpdos) > 0:
                module['TxPdo'] = txpdos
            modules.append(module)

        return {'Module': modules}

    def model_to_module_rxpdo(self, module):
        pdos = []
        for p in module.ethercat.rxpdo:
            pdo = {
                '@Fixed': True,
                '@Sm': 2,
                'Index': {
                    '@DependOnSlot': True,
                    '$': p.index
                },
                'Name': p.name,
                'Entry': self.model_to_pdo_entries(module, module.outputs, p)
            }
            pdos.append(pdo)
        return pdos

    def model_to_module_txpdo(self, module):
        pdos = []
        for p in module.ethercat.txpdo:
            pdo = {
                '@Fixed': True,
                '@Sm': 3,
                'Index': {
                    '@DependOnSlot': True,
                    '$': p.index
                },
                'Name': p.name,
                'Entry': self.model_to_pdo_entries(module, module.inputs, p)
            }
            pdos.append(pdo)
        return pdos

    def model_to_pdo_entries(self, module, signals, pdo):
        entries = []
        for e in pdo.entries:
            signal = module.get_signal(signals, e.signal)
            if not signal.is_array:
                entry = {
                    'Index': e.index,
                    'SubIndex': e.subindex,
                    'BitLen': signal.bitlen,
                    'Name': signal.name,
                    'DataType': ecat_type[signal.datatype]
                }
                entries.append(entry)
            else:
                for subindex in range (1, signal.array_length + 1):
                    entry = {
                        'Index': e.index,
                        'SubIndex': str(subindex),
                        'BitLen': ecat_bitlen[ecat_type[signal.datatype]],
                        'Name': signal.name,
                        'DataType': ecat_type[signal.datatype]
                    }
                    entries.append(entry)
        return entries

    def model_to_directory(self, device):
        directory = {
            'DataTypes': {
                'DataType': [
                    self.dt1018(),
                    self.dt10f3(),
                    self.dtF000(),
                ]
            },
            'Objects': {
                'Object': [
                    self.x1000_device_type(device),
                    self.x1008_device_name(device),
                    self.x1009_hardware_version(device),
                    self.x100A_software_version(device),
                    self.x1018_identity(device),
                    self.x10F3_diagnosis_history(device),
                    self.x10F8_timestamp_object(device),
                    self.xF000_module_device_profile(device),
                    self.xF010_module_profile_list(device),
                    self.xF050_module_detected_list(device),
                    self.x1C12_sync_manager_2(device),
                    self.x1C13_sync_manager_3(device),
                    self.x1C00_sync_manager_comm_type(device)
                ]
            }
        }

        for name in self.data_types:
            directory['DataTypes']['DataType'].append(self.data_types[name])

        return directory

    def model_diag_messages (self, device):
        diags = {'DiagMessage':[]}

        for slot in device.slots:
            module = self.model.get_module(slot.module)
            for alarm in module.alarms:
                diag = {
                        'TextId': alarm.error_code,
                        'MessageText': alarm.message
                   }
                diags['DiagMessage'].append(diag)
        return diags

    def add_data_type(self, data_type):
        type_name = data_type['Name']
        if type_name not in self.data_types.keys():
            self.data_types[type_name] = data_type
        return type_name

    def add_array_type(self, name, base_type, length):
        type = {
            'Name': name,
            'BaseType': base_type,
            'BitSize': self.data_types[base_type]['BitSize'] * length,
            'ArrayInfo': {
                'LBound': 1,
                'Elements': length
            }
        }
        return self.add_data_type(type)

    def add_string_type(self, string):
        str_type = {
            'Name': f'STRING({len(string)})',
            'BitSize': len(string) * 8
        }
        return self.add_data_type(str_type)

    def add_DT1C00_type(self):
        type = {
            'Name': 'DT1C00',
            'BitSize': 0,
            'SubItem': [
                {
                    'SubIdx': '0',
                    'Name': 'Max SubIndex',
                    'Type': 'USINT',
                    'BitSize': 8,
                    'BitOffs': 0,
                    'Flags': {
                        'Access': 'ro'
                    }
                },
                {
                    'Name': 'Elements',
                    'Type': self.add_array_type('DT1C00ARR', 'USINT', 4),
                    'BitSize': self.data_types['DT1C00ARR']['BitSize'],
                    'BitOffs': 16,
                    'Flags': {
                        'Access': 'ro'
                    }
                }
            ]
        }
        type['BitSize'] = self.data_types['DT1C00ARR']['BitSize'] + 16
        return self.add_data_type(type)

    def add_DTF010_type(self, modules):
        type = {
            'Name': 'DTF010',
            'BitSize': 0,
            'SubItem': [
                {
                    'SubIdx': '0',
                    'Name': 'Max SubIndex',
                    'Type': 'USINT',
                    'BitSize': 8,
                    'BitOffs': 0,
                    'Flags': {
                        'Access': 'ro'
                    }
                },
                {
                    'Name': 'Elements',
                    'Type': self.add_array_type('DTF010ARR', 'UDINT', len(modules)),
                    'BitSize': self.data_types['DTF010ARR']['BitSize'],
                    'BitOffs': 16,
                    'Flags': {
                        'Access': 'ro'
                    }
                }
            ]
        }
        type['BitSize'] = self.data_types['DTF010ARR']['BitSize'] + 16
        return self.add_data_type(type)

    def add_DTF050_type(self, modules):
        type = {
            'Name': 'DTF050',
            'BitSize': 0,
            'SubItem': [
                {
                    'SubIdx': '0',
                    'Name': 'Max SubIndex',
                    'Type': 'USINT',
                    'BitSize': 8,
                    'BitOffs': 0,
                    'Flags': {
                        'Access': 'ro'
                    }
                },
                {
                    'Name': 'Elements',
                    'Type': self.add_array_type('DTF050ARR', 'UDINT', len(modules)),
                    'BitSize': self.data_types['DTF050ARR']['BitSize'],
                    'BitOffs': 16,
                    'Flags': {
                        'Access': 'ro'
                    }
                }
            ]
        }
        type['BitSize'] = self.data_types['DTF050ARR']['BitSize'] + 16
        return self.add_data_type(type)

    def add_DT1C12_type(self, slots):
        n_rxpdos = 0
        for ix, slot in enumerate(slots):
            m = self.model.get_module(slot.module)
            rxpdo = self.model_to_module_rxpdo(m)
            if len(rxpdo) > 0:
                n_rxpdos = n_rxpdos + 1

        type = {
            'Name': 'DT1C12',
            'BitSize': 0,
            'SubItem': [
                {
                    'SubIdx': '0',
                    'Name': 'Max SubIndex',
                    'Type': 'USINT',
                    'BitSize': 8,
                    'BitOffs': 0,
                    'Flags': {
                        'Access': 'ro'
                    }
                },
                {
                    'Name': 'Elements',
                    'Type': self.add_array_type('DT1C12ARR', 'UINT', n_rxpdos),
                    'BitSize': self.data_types['DT1C12ARR']['BitSize'],
                    'BitOffs': 16,
                    'Flags': {
                        'Access': 'ro'
                    }
                }
            ]
        }
        type['BitSize'] = self.data_types['DT1C12ARR']['BitSize'] + 16
        return self.add_data_type(type)

    def add_DT1C13_type(self, slots):
        n_txpdos = 0
        for ix, slot in enumerate(slots):
            m = self.model.get_module(slot.module)
            txpdo = self.model_to_module_txpdo(m)
            if len(txpdo) > 0:
                n_txpdos = n_txpdos + 1

        type = {
            'Name': 'DT1C13',
            'BitSize': 0,
            'SubItem': [
                {
                    'SubIdx': '0',
                    'Name': 'Max SubIndex',
                    'Type': 'USINT',
                    'BitSize': 8,
                    'BitOffs': 0,
                    'Flags': {
                        'Access': 'ro'
                    }
                },
                {
                    'Name': 'Elements',
                    'Type': self.add_array_type('DT1C13ARR', 'UINT', n_txpdos),
                    'BitSize': self.data_types['DT1C13ARR']['BitSize'],
                    'BitOffs': 16,
                    'Flags': {
                        'Access': 'ro'
                    }
                }
            ]
        }
        type['BitSize'] = self.data_types['DT1C13ARR']['BitSize'] + 16
        return self.add_data_type(type)

    def DTF010_subitems(self, modules):
        subitems = [{
            'Name': 'Max SubIndex',
            'Info': {
                'DefaultValue': str(len(modules))
            }
        }]
        for m in modules:
            subitems.append({
                'Name': m.name,
                'Info': {
                    'DefaultValue': hexify(m.ethercat.profile)
                }
            })
        return subitems

    def DTF050_subitems(self, modules):
        subitems = [{
            'Name': 'Max SubIndex',
            'Info': {
                'DefaultValue': str(len(modules))
            }
        }]
        for ix, m in enumerate(modules):
            subitems.append({
                'Name': m.name,
                'Info': {
                    'DefaultValue': hexify(f'0x{(ix+1):04x}')
                }
            })
        return subitems

    def add_DT1C12_subitems(self, device):
        subitems = []
        pdos = []

        for ix, slot in enumerate(device.slots):
            m = self.model.get_module(slot.module)
            pdo = self.model_to_module_rxpdo(m)
            if len(pdo) > 0:
                pdos.append(0x1600 + slot_pdo_increment * ix)

        max_subindex = {
            'Name': 'Max SubIndex',
            'Info': {
                'DefaultValue': str(len(pdos))
            }
        }

        subitems.append(max_subindex)

        for pdo in pdos:
            subitem = {
                'Name': 'PDO Mapping',
                'Info': {
                    'DefaultValue': hexify(f'0x{pdo:04x}')
                }
            }
            subitems.append(subitem)

        return subitems

    def add_DT1C13_subitems(self, device):
        subitems = []
        pdos = []

        for ix, slot in enumerate(device.slots):
            m = self.model.get_module(slot.module)
            pdo = self.model_to_module_txpdo(m)
            if len(pdo) > 0:
                pdos.append(0x1A00 + slot_pdo_increment * ix)

        max_subindex = {
            'Name': 'Max SubIndex',
            'Info': {
                'DefaultValue': str(len(pdos))
            }
        }

        subitems.append(max_subindex)

        for pdo in pdos:
            subitem = {
                'Name': 'PDO Mapping',
                'Info': {
                    'DefaultValue': hexify(f'0x{pdo:04x}')
                }
            }
            subitems.append(subitem)

        return subitems

    def dt1018(self):
        return {
            'Name': 'DT1018',
            'BitSize': 144,
            'SubItem': [
                {
                    'SubIdx': '0',
                    'Name': 'Max SubIndex',
                    'Type': 'USINT',
                    'BitSize': 8,
                    'BitOffs': 0,
                    'Flags': {
                        'Access': 'ro'
                    }
                },
                {
                    'SubIdx': '1',
                    'Name': 'Vendor ID',
                    'Type': 'UDINT',
                    'BitSize': 32,
                    'BitOffs': 16,
                    'Flags': {
                        'Access': 'ro'
                    }
                },
                {
                    'SubIdx': '2',
                    'Name': 'Product Code',
                    'Type': 'UDINT',
                    'BitSize': 32,
                    'BitOffs': 48,
                    'Flags': {
                        'Access': 'ro'
                    }
                },
                {
                    'SubIdx': '3',
                    'Name': 'Revision Number',
                    'Type': 'UDINT',
                    'BitSize': 32,
                    'BitOffs': 80,
                    'Flags': {
                        'Access': 'ro'
                    }
                },
                {
                    'SubIdx': '4',
                    'Name': 'Serial Number',
                    'Type': 'UDINT',
                    'BitSize': 32,
                    'BitOffs': 112,
                    'Flags': {
                        'Access': 'ro'
                    }
                }
            ]
        }
    def dt10f3(self):
        dt = {
            'Name': 'DT10F3',
            'BitSize':  64 + 16*128 ,
            'SubItem': [
                {
                    'SubIdx': '0',
                    'Name': 'Max SubIndex',
                    'Type': 'USINT',
                    'BitSize': 8,
                    'BitOffs': 0,
                    'Flags': {
                        'Access': 'ro'
                    }
                },
                {
                    'SubIdx': '1',
                    'Name': 'Maximum Messages',
                    'Type': 'USINT',
                    'BitSize': 8,
                    'BitOffs': 16,
                    'Flags': {
                        'Access': 'ro'
                    }
                },
                {
                    'SubIdx': '2',
                    'Name': 'Newest Message',
                    'Type': 'USINT',
                    'BitSize': 8,
                    'BitOffs': 24,
                    'Flags': {
                        'Access': 'ro'
                    }
                },
                {
                    'SubIdx': '3',
                    'Name': 'Newest Acknowledged Message',
                    'Type': 'USINT',
                    'BitSize': 8,
                    'BitOffs': 32,
                    'Flags': {
                        'Access': 'rw'
                    }
                },
                {
                    'SubIdx': '4',
                    'Name': 'New Messages Available',
                    'Type': 'BOOL',
                    'BitSize': 1,
                    'BitOffs': 40,
                    'Flags': {
                        'Access': 'ro'
                    }
                },
                {
                    'SubIdx': '5',
                    'Name': 'Flags',
                    'Type': 'UINT',
                    'BitSize': 16,
                    'BitOffs': 48,
                    'Flags': {
                        'Access': 'rw'
                    }
                }
            ]
        }

        for i in range (16):
            msg = {
                    'SubIdx': str(6 + i),
                    'Name': f'Diagnosis Message {1+i}',
                    'Type': 'ARRAY [0..15] OF BYTE',
                    'BitSize': 128,
                    'BitOffs': 64 + i*128,
                    'Flags': {
                        'Access': 'ro'
                    }
            }
            dt['SubItem'].append(msg)

        return dt


    def dtF000(self):
        return {
            'Name': 'DTF000',
            'BitSize': 48,
            'SubItem': [
                {
                    'SubIdx': '0',
                    'Name': 'Max SubIndex',
                    'Type': 'USINT',
                    'BitSize': 8,
                    'BitOffs': 0,
                    'Flags': {
                        'Access': 'ro'
                    }
                },
                {
                    'SubIdx': '1',
                    'Name': 'Index distance',
                    'Type': 'UINT',
                    'BitSize': 16,
                    'BitOffs': 16,
                    'Flags': {
                        'Access': 'ro'
                    }
                },
                {
                    'SubIdx': '2',
                    'Name': 'Maximum number of modules',
                    'Type': 'UINT',
                    'BitSize': 16,
                    'BitOffs': 32,
                    'Flags': {
                        'Access': 'ro'
                    }
                }
            ]
        }

    def x1000_device_type(self, device):
        return {
            'Index': '#x1000',
            'Name': 'Device Type',
            'Type': 'UDINT',
            'BitSize': 32,
            'Info': {
                'DefaultValue': hexify(device.ethercat.profile)
            },
            'Flags': {
                'Access': 'ro',
                'Category': 'm'
            }
        }

    def x1008_device_name(self, device):
        return {
            'Index': '#x1008',
            'Name': 'Device Name',
            'Type': self.add_string_type(device.name),
            'BitSize': len(device.name) * 8,
            'Info': {
                'DefaultString': device.name
            },
            'Flags': {
                'Access': 'ro'
            }
        }

    def x1009_hardware_version(self, device):
        return {
            'Index': '#x1009',
            'Name': 'Hardware Version',
            'Type': self.add_string_type(device.hardware_release),
            'BitSize': len(device.hardware_release) * 8,
            'Info': {
                'DefaultString': device.hardware_release
            },
            'Flags': {
                'Access': 'ro',
                'Category': 'o'
            }
        }

    def x100A_software_version(self, device):
        return {
            'Index': '#x100A',
            'Name': 'Software Version',
            'Type': self.add_string_type(device.software_release),
            'BitSize': len(device.software_release) * 8,
            'Info': {
                'DefaultString': device.software_release
            },
            'Flags': {
                'Access': 'ro'
            }
        }

    def x1018_identity(self, device):
        return {
            'Index': '#x1018',
            'Name': 'Identity Object',
            'Type': 'DT1018',
            'BitSize': 144,
            'Info': {
                'SubItem': [
                    {
                        'Name': 'Max SubIndex',
                        'Info': {
                            'DefaultValue': '4'
                        }
                    },
                    {
                        'Name': 'Vendor ID',
                        'Info': {
                            'DefaultValue': hexify(self.model.ethercat.vendor_id)
                        }
                    },
                    {
                        'Name': 'Product Code',
                        'Info': {
                            'DefaultValue': hexify(device.ethercat.product_code)
                        }
                    },
                    {
                        'Name': 'Revision Number',
                        'Info': {
                            'DefaultValue': device.ethercat.revision
                        }
                    },
                    {
                        'Name': 'Serial Number',
                        'Info': {
                            'DefaultValue': '#x00000000'
                        }
                    }
                ]
            },
            'Flags': {
                'Access': 'ro'
            }
        }

    def x10F3_diagnosis_history(self, device):
        return {
            'Index': '#x10F3',
            'Name': 'Diagnosis History',
            'Type': 'DT10F3',
            'BitSize': 64 + 16*128,
            'Info': {
                'SubItem':[
                    {
                        'Name': 'Max SubIndex',
                        'Info': {
                            'DefaultValue': '#x15'
                        }
                    }

                ]
            },
            'Flags': {
                'Access': 'ro'
            }
        }
    def x10F8_timestamp_object(self, device):
        return {
            'Index': '#x10F8',
            'Name': 'Timestamp Object',
            'Type': 'ULINT',
            'BitSize': 64,
            'Info': {
                'Unit': '#xf7030000'
                },
            'Flags': {
                'Access': 'ro'
            }
        }

    def xF000_module_device_profile(self, device):
        return {
            'Index': '#xF000',
            'Name': 'Module Device Profile',
            'Type': 'DTF000',
            'BitSize': 48,
            'Info': {
                'SubItem': [
                    {
                        'Name': 'Max SubIndex',
                        'Info': {
                            'DefaultValue': '2'
                        }
                    },
                    {
                        'Name': 'Index distance',
                        'Info': {
                            'DefaultValue': '#x800'
                        }
                    },
                    {
                        'Name': 'Maximum number of modules',
                        'Info': {
                            'DefaultValue': str(len(self.model.modules))
                        }
                    }
                ]
            },
            'Flags': {
                'Access': 'ro'
            }
        }

    def xF010_module_profile_list(self, device):
        return {
            'Index': '#xF010',
            'Name': 'Module Profile List',
            'Type': self.add_DTF010_type(self.model.modules),
            'BitSize': self.data_types['DTF010']['BitSize'],
            'Info': {
                'SubItem': self.DTF010_subitems(self.model.modules)
            },
            'Flags': {
                'Access': 'ro'
            }
        }

    def xF050_module_detected_list(self, device):
        return {
            'Index': '#xF050',
            'Name': 'Module Detected List',
            'Type': self.add_DTF050_type(self.model.modules),
            'BitSize': self.data_types['DTF050']['BitSize'],
            'Info': {
                'SubItem': self.DTF050_subitems(self.model.modules)
            },
            'Flags': {
                'Access': 'ro'
            }
        }

    def x1C12_sync_manager_2(self, device):
        return {
            'Index': '#x1C12',
            'Name': 'Sync Manager 2 PDO Assignment',
            'Type': self.add_DT1C12_type(device.slots),
            'BitSize': self.data_types['DT1C12']['BitSize'],
            'Info': {
                'SubItem': self.add_DT1C12_subitems(device)
            },
            'Flags': {
                'Access': 'ro'
            }
        }

    def x1C13_sync_manager_3(self, device):
        return {
            'Index': '#x1C13',
            'Name': 'Sync Manager 3 PDO Assignment',
            'Type': self.add_DT1C13_type(device.slots),
            'BitSize': self.data_types['DT1C13']['BitSize'],
            'Info': {
                'SubItem': self.add_DT1C13_subitems(device)
            },
            'Flags': {
                'Access': 'ro'
            }
        }

    def x1C00_sync_manager_comm_type(self, device):
        return {
            'Index': '#x1C00',
            'Name': 'Sync Manager Communication Type',
            'Type': self.add_DT1C00_type(),
            'BitSize': self.data_types['DT1C00']['BitSize'],
            'Info': {
                'SubItem': [
                    {
                        'Name': 'Max SubIndex',
                        'Info': {
                            'DefaultValue': '4'
                        }
                    },
                    {
                        'Name': 'Communications Type SM0',
                        'Info': {
                            'DefaultValue': '1'
                        }
                    },
                    {
                        'Name': 'Communications Type SM1',
                        'Info': {
                            'DefaultValue': '2'
                        }
                    },
                    {
                        'Name': 'Communications Type SM2',
                        'Info': {
                            'DefaultValue': '3'
                        }
                    },
                    {
                        'Name': 'Communications Type SM3',
                        'Info': {
                            'DefaultValue': '4'
                        }
                    }
                ]
            },
            'Flags': {
                'Access': 'ro'
            }
        }
